# -*- coding: utf-8 -*-
# Copyright (c) 2016 - 2020 Sqreen. All rights reserved.
# Please refer to our terms for more information:
#
#     https://www.sqreen.io/terms.html
#

"""Radix tree implementation optimized to process IPs numerically"""

import struct
import sys
from logging import getLogger

from sqreen.utils import ip_address, ip_network

from ._vendors.ipaddress import (
    IPv4Address,
    IPv4Network,
    IPv6Address,
    IPv6Network,
)

IS_PYTHON_2 = sys.version_info.major == 2
LOGGER = getLogger(__name__)


class Node(object):
    """Node of the tree"""

    def __init__(self, bit_width):
        """Initialize the node to its default value"""
        self.value = 0
        self.depth = 0
        self.length = 0
        self.bit_width = bit_width
        self.is_network = False
        self.node0 = None
        self.node1 = None
        self.payload = None

    def set_root(self, ip, is_network):
        """Configure a node as the root node of the radix tree"""
        if ip.version == 4:
            self.bit_width = 32
        else:
            self.bit_width = 128

        self.value = Radix.get_value(ip, is_network)
        self.depth = 0
        self.is_network = is_network
        if is_network:
            self.length = ip.prefixlen
        else:
            self.length = self.bit_width

    def is_empty_root(self):
        """Return whether this node is the root node and is empty (which is different from a single root node containing a all-zero IP"""
        return self.depth == 0 and self.length == 0 and self.node0 is None

    def set_value(self, ip_value, length_to_insert, is_network):
        """Small method setting the value of a node, only picking a few bits from an overall value"""
        self.value = ip_value
        self.value &= (1 << length_to_insert) - 1

        self.depth = self.bit_width - length_to_insert
        self.length = length_to_insert
        self.is_network = is_network

    def get_polarity(self):
        """Return the highest relevant bit (according to the depth and length of the node) is"""
        return (self.value >> (self.bit_width - 1 - self.depth)) & 1

    def clamp_value(self):
        """Constraint the value of a node to the relevant range, according to its depth and length.
        Make shenanigans with the node position easier to pull off, especially considering the bit magic involved
        """
        upper_clamp = self.bit_width - self.depth
        lower_clamp = upper_clamp - self.length

        self.value &= ((1 << upper_clamp) - 1)
        self.value &= ~((1 << lower_clamp) - 1)

    def restrict_length(self, new_length, other_new_child):
        """Reduce the length of the node, turning what is left in a child of this node"""
        length_to_remove = self.length - new_length
        if new_length <= 0 or length_to_remove <= 0:
            LOGGER.error("Radix Tree internal inconsistency: restrict length called with new_length %d with length_to_remove %d", new_length, length_to_remove)
            return False

        if self.is_network:
            network_offset = self.bit_width - self.depth - self.length
        else:
            network_offset = 0

        # Create a new child with the section we have to remove, assuming we even want to keep it
        new_child = Node(self.bit_width)
        new_child.value = self.value
        new_child.payload = self.payload
        new_child.is_network = self.is_network
        new_child.length = length_to_remove + network_offset
        new_child.depth = self.depth + new_length
        new_child.clamp_value()
        new_child.length -= network_offset

        if self.node0 is not None or self.node1 is not None:
            if self.is_network:
                LOGGER.error("Radix Tree internal inconsistency: IP ranges are not supposed to have children! %s", self)
                return False

            new_child.node0 = self.node0
            new_child.node1 = self.node1

        if new_child.get_polarity() == other_new_child.get_polarity():
            LOGGER.warning("Radix Tree internal inconsistency: we likely made a mistake computing the length to remove from a node")
            return False

        # We now need to determine the polarity of new_child to know to which node[0-1] we need to set it to
        if new_child.get_polarity():
            self.node0 = other_new_child
            self.node1 = new_child
        else:
            self.node0 = new_child
            self.node1 = other_new_child

        # Remove a section at our end
        self.length = new_length
        self.payload = None
        self.clamp_value()
        self.is_network = False
        return True

    def remove_child_with_polarity(self, polarity):
        """Remove one of our child and merge the orphan with us"""
        if self.node0 is None or self.node1 is None:
            LOGGER.warning("Radix Tree internal inconsistency: trying to remove a child from a node without children")
            return

        # We're removing node1, so we have to merge with node0
        if polarity:
            worthy_child = self.node0
        else:
            worthy_child = self.node1

        self.length += worthy_child.length
        self.value |= worthy_child.value
        self.is_network = worthy_child.is_network
        self.payload = worthy_child.payload
        self.node0 = worthy_child.node0
        self.node1 = worthy_child.node1

    def all_payloads(self):
        if self.payload:
            yield self.payload

        if self.node0:
            for p in self.node0.all_payloads():
                yield p

        if self.node1:
            for p in self.node1.all_payloads():
                yield p

    def __str__(self, parent_value=0):
        """Recursively generate the string representation of the node and its children,
           by passing the values of the higher bits"""

        if self.is_empty_root():
            return ""

        parent_value |= self.value

        if self.node0 is not None:
            output = self.node0.__str__(parent_value) + ", " + self.node1.__str__(parent_value)
        else:
            if self.bit_width == 32:
                output = str(IPv4Address(parent_value))
            else:
                output = str(IPv6Address(parent_value))

            if self.is_network:
                mask_width = self.depth + self.length
                output += "/" + str(mask_width)

        return output


class Radix(object):
    """Main data structure of the Radix tree. Contains two trees: one for IPv4, one for IPv6"""
    def __init__(self, ip, is_network=False):
        """Initialize the two tries"""
        self.root4 = Node(32)
        self.root6 = Node(128)

        ip = self.enforce_ip(ip, is_network)
        if ip is not None:
            if ip.version == 4:
                self.root4.set_root(ip, is_network)
            else:
                self.root6.set_root(ip, is_network)

            self.empty = False
        else:
            self.empty = True

    @staticmethod
    def create_ip(ip, is_network):
        """Convert, if necessary, the input IP into the appropriate object"""
        if ip is None:
            return None

        if is_network:
            if isinstance(ip, IPv4Network) or isinstance(ip, IPv6Network):
                return ip
            if IS_PYTHON_2 and isinstance(ip, str):
                ip = ip.decode("utf-8")
            try:
                return ip_network(ip, False)
            except ValueError:
                return None

        else:
            if isinstance(ip, IPv4Address) or isinstance(ip, IPv6Address):
                return ip
            if IS_PYTHON_2 and isinstance(ip, str):
                ip = ip.decode("utf-8")
            try:
                return ip_address(ip)
            except ValueError:
                return None

    @classmethod
    def enforce_ip(cls, ip, is_network):
        """Wrap around create_ip and turn IPv4 addresses mapped in IPv6 back into IPv4"""
        ip_object = cls.create_ip(ip, is_network)

        # IPv6 addresses may actually be hidden IPv4 addresses. We want to detect them and move them to the proper tree
        if ip_object is None or ip_object.version != 6:
            return ip_object

        ip_value = cls.get_value(ip_object, is_network)
        ip_value_without_ipv4 = ip_value >> 32
        # The two approaches are ::0000:IPv4 (RFC4291 @ 2.5.5.1, deprecated) and ::ffff::IPv4 (RFC4291 @ 2.5.5.2)
        if ip_value_without_ipv4 == 0x0000 or ip_value_without_ipv4 == 0xffff:
            # Clear the bits higher of the numerical value, then create a new object (which will be IPv4 as the numerical value is low enough)
            ip_value &= 0xffffffff
            if not is_network:
                return IPv4Address(ip_value)

            # If this is a network, we have to check if the subnet is small enough
            if ip_object.prefixlen >= 96:
                return IPv4Network(ip_value)

        return ip_object

    def exist(self, ip, is_network=False):
        """Crawl the appropriate tree in order to check whether the IP represented by the object exist in the tree"""
        if self.empty:
            return None

        target_value = self.get_value(ip, is_network)
        if ip.version == 4:
            cur_node = self.root4
        else:
            cur_node = self.root6

        # Is the tree empty?
        if cur_node.node0 is None and cur_node.node1 is None and cur_node.length == 0:
            return None

        bit_width = cur_node.bit_width
        search_length_left = bit_width

        # The following algorithm is optimized to minimize the number of branches
        while True:

            # We check if the cur_node and the target share the same value.
            # This will however compare the whole address, instead of only the relevant section
            combined_value = target_value ^ cur_node.value

            # Mark the bits we just compared as done
            search_length_left -= cur_node.length

            # This will remove any irrelevant low bits from the compare result
            combined_value >>= search_length_left

            # The IP doesn't match the current node: the match fail
            values_match = combined_value == 0

            # We fully checked the IP. This would mean that assuming the values matched, the full IP match
            # If this node is a network we blocked, there is no need to go any further
            checked_full_ip = search_length_left == 0 or cur_node.is_network

            if not values_match or checked_full_ip:
                # If we're checking a network, the node we're matching to need to also be one
                if is_network:
                    values_match &= cur_node.is_network and (cur_node.depth + cur_node.length <= ip.prefixlen)

                if values_match:
                    return self.string_for_match(ip, cur_node, is_network)

                return None

            # This simply create a mask, removing any irrelevant bit
            target_value &= (1 << search_length_left) - 1

            # We pick the most significant bit
            highest_bit = target_value >> (search_length_left - 1)

            if highest_bit == 0:
                cur_node = cur_node.node0
            else:
                cur_node = cur_node.node1

    def match(self, ip, is_network=False):
        """Call .exist to check if an IP/network exist in the tree, but is more liberal in the input it will accept"""
        ip = self.enforce_ip(ip, is_network)
        if ip is None:
            return None

        return self.exist(ip, is_network)

    def insert(self, ip, is_network=False, payload=None):
        """Insert an IP/network in the tree"""
        ip = self.enforce_ip(ip, is_network)
        if ip is None:
            return False

        if ip.version == 4:
            cur_root = self.root4
        else:
            cur_root = self.root6

        cur_node = cur_root

        if self.empty or cur_root.is_empty_root():
            self.empty = False
            cur_node.set_root(ip, is_network)
            cur_node.payload = payload
            return True

        target_value = self.get_value(ip, is_network)
        search_length_left = cur_node.bit_width

        if self.exist(ip, is_network) is not None:
            return False

        # The following algorithm is based on `exist`, although duplicated in order to avoid adding anything to the hot path
        while True:

            combined_value = target_value ^ cur_node.value
            node_offset = search_length_left - cur_node.length
            combined_value >>= node_offset

            hit_end_network_mask = is_network and (cur_node.length + cur_node.depth >= ip.prefixlen)

            # Have we found the divergence point?
            if combined_value != 0 or hit_end_network_mask:

                # If a network, we remove the bits within the network mask
                if hit_end_network_mask:
                    network_mask_bits = (cur_node.length + cur_node.depth) - ip.prefixlen
                    node_offset += network_mask_bits
                    combined_value >>= network_mask_bits
                else:
                    network_mask_bits = 0

                # We compute the number of mismatched bits by progressively shifting the compare
                number_of_mismatched_bits = 0
                while (combined_value >> number_of_mismatched_bits) != 0:
                    number_of_mismatched_bits += 1

                # If we're adding a network, and we match with something within our mask,
                # all we really have to do is the shorten the node to only contain the network IP, remove any node and turn into a network
                if is_network and number_of_mismatched_bits == 0:
                    # We need to restrict the node's length to prefixlen. We make sure that this is possible. Otherwise, we bail out
                    if ip.prefixlen < cur_node.depth:
                        LOGGER.warning("Radix Tree internal inconsistency: We went too deep in the tree before inserting %s, which isn't supposed to be possible.", ip)
                        return False

                    cur_node.length = ip.prefixlen - cur_node.depth
                    cur_node.clamp_value()
                    cur_node.node0 = None
                    cur_node.node1 = None
                    cur_node.is_network = True
                    cur_node.payload = payload

                else:
                    new_node = Node(cur_node.bit_width)
                    new_node.set_value(self.get_value(ip, is_network), number_of_mismatched_bits + node_offset, is_network)
                    new_node.payload = payload

                    # Networks ignore a part of their address, this need to be reflected in their length. No need to update the value as it's already to 0
                    if is_network:
                        new_node.length = ip.prefixlen - new_node.depth

                    is_node_root = cur_node.depth == 0 and cur_node.length == cur_root.length

                    # Do we need to split the root node? (the last condition can only be false is something is going really wrong with .depth)
                    if not is_node_root or search_length_left != cur_node.bit_width:
                        # Nope! Awesome!
                        return cur_node.restrict_length(cur_node.length - number_of_mismatched_bits - network_mask_bits, new_node)

                    new_root = Node(cur_node.bit_width)
                    new_root.set_value(cur_node.value, cur_node.bit_width, False)
                    new_root.length = new_node.depth
                    # We add the full value to the root node, and we then need to clamp it
                    new_root.clamp_value()

                    cur_node.depth = new_node.depth
                    cur_node.length -= new_node.depth
                    cur_node.clamp_value()

                    # We check the polarity of the new node
                    if new_node.get_polarity():
                        new_root.node0 = cur_node
                        new_root.node1 = new_node
                    else:
                        new_root.node0 = new_node
                        new_root.node1 = cur_node

                    if ip.version == 4:
                        self.root4 = new_root
                    else:
                        self.root6 = new_root

                return True

            search_length_left -= cur_node.length
            if search_length_left == 0:
                return False

            if cur_node.is_network:
                return False

            target_value &= (1 << search_length_left) - 1
            highest_bit = target_value >> (search_length_left - 1)

            if highest_bit == 0:
                cur_node = cur_node.node0
            else:
                cur_node = cur_node.node1

    def remove(self, ip, is_network=False):
        """Remove an IP/network from the tree"""

        ip = self.enforce_ip(ip, is_network)
        if ip is None:
            return False

        if self.exist(ip, is_network) is None:
            return True

        if ip.version == 4:
            cur_node = self.root4
        else:
            cur_node = self.root6

        # Only one node exist and so are we. This mean we need to dump the root node. Ugh...
        if cur_node.length == cur_node.bit_width or cur_node.is_network:

            # Removing a smaller network or an ip from a network is not allowed
            if (is_network and ip.prefixlen > cur_node.depth + cur_node.length)\
                    or (not is_network and cur_node.is_network):
                return False

            cur_node.length = 0
            cur_node.node0 = cur_node.node1 = cur_node.payload = None

            if ip.version == 4:
                other_node = self.root6
            else:
                other_node = self.root4

            if other_node.is_empty_root():
                self.empty = True

            return True

        target_value = self.get_value(ip, is_network)
        current_offset = cur_node.bit_width - 1

        while current_offset > 0:

            # The next node is down the node1 path
            next_polarity = (target_value >> (current_offset - cur_node.length)) & 1
            if next_polarity:
                next_node = cur_node.node1
            else:
                next_node = cur_node.node0

            # We advance the current_offset
            current_offset -= cur_node.length

            # We ran in a network. This is a bit of a weird case where some cases are not allowed
            if next_node.is_network:
                # This is a network, and we're trying to extract a single IP which isn't supported yet
                if not is_network:
                    return False

                # If the network we're trying to remove is a subnet of this network, this is again not supported
                if next_node.depth + next_node.length < ip.prefixlen:
                    return False

                # Network nodes can't have children, meaning that the next call will trigger
                if next_node.node0 is not None or next_node.node1 is not None:
                    LOGGER.warning("Radix Tree internal inconsistency: found a network node with children")
                    next_node.node0 = next_node.node1 = None

            # Is the next node hitting the bottom? If so, it doesn't have children and we found the need we wanted to delete
            if next_node.node0 is None:
                cur_node.remove_child_with_polarity(next_polarity)
                return True

            # We proceed
            cur_node = next_node

        return False

    def clear(self):
        """Clear the tree"""
        if not self.empty:
            self.empty = True
            self.root4 = Node(self.root4.bit_width)
            self.root6 = Node(self.root6.bit_width)

    def all_payloads(self):
        for p in self.root4.all_payloads():
            yield p

        for p in self.root6.all_payloads():
            yield p

    @staticmethod
    def get_value(ip, is_network):
        """Return an integer containing the numerical value of the bit representation of an IP"""
        if is_network:
            target_array = ip.network_address.packed
        else:
            target_array = ip.packed

        if ip.version == 4:
            if IS_PYTHON_2:
                return struct.unpack('>I', target_array)[0]

            return (target_array[0] << 24) | (target_array[1] << 16) | (target_array[2] << 8) | target_array[3]
        else:
            if IS_PYTHON_2:
                target_array = struct.unpack('>QQ', target_array)
                return target_array[0] << 64 | target_array[1]

            output = 0
            for ip_item in target_array:
                output <<= 8
                output |= ip_item

            return output

    @classmethod
    def string_for_match(cls, ip, node, ip_is_network):
        """Return the string corresponding to IP we matched from the tree (i.e. the rule we matched)"""
        # Do we have a specific payload to return?
        if node.payload is not None:
            return node.payload

        if not node.is_network:
            return str(ip)

        # Get the network IP's, based on the target's
        full_ip_value = cls.get_value(ip, ip_is_network)

        relevant_length = node.depth + node.length
        length_to_trim = node.bit_width - relevant_length
        if length_to_trim != 0:
            full_ip_value &= ~((1 << length_to_trim) - 1)

        # IPv4
        if node.bit_width == 32:
            return str(IPv4Address(full_ip_value)) + "/" + str(relevant_length)

        # IPv6
        return str(IPv6Address(full_ip_value)) + "/" + str(relevant_length)

    def __repr__(self):
        return "{}({!r})".format(self.__class__.__name__, str(self))

    def __str__(self):
        if self.empty:
            return ""

        radix_v4 = str(self.root4)
        radix_v6 = str(self.root6)

        if radix_v4 == "":
            return radix_v6

        elif radix_v6 == "":
            return radix_v4

        return radix_v4 + ", " + radix_v6



# -*- coding: utf-8 -*-
# Copyright (c) 2016 - 2020 Sqreen. All rights reserved.
# Please refer to our terms for more information:
#
#     https://www.sqreen.io/terms.html
#
"""Provide task-contextual storage for asyncio.

In order to inherit contexts when a sub-task is spawned (e.g. when calling
asyncio.gather), it is required to set up a custom task factory in the event
loop. This task factory will attach a shallow copy of the current context to
every new task.
"""

from asyncio import Task
from functools import wraps
from logging import getLogger

LOGGER = getLogger(__name__)


_TASK_CONTEXT_ATTR = "_sqreen_context"


def with_current_task_context(func):
    """Call the decorated function with the current task context.

    The task context is passed as an additional first argument. If no task is
    running or the context was not set, the function is not called and an error
    message is issued.
    """
    @wraps(func)
    def wrapper(*args, **kwargs):
        task = Task.current_task()
        if task is None:
            # TODO This log message is displayed at startup because we set a
            # request_recorder at initialization, before the event loop is
            # started. Let's dismiss it for now.
            LOGGER.info("No current task, is the event loop running?")
            return
        context = getattr(task, _TASK_CONTEXT_ATTR, None)
        if context is None:
            LOGGER.error(
                "No context attached to the current task, "
                "is the task factory correctly set?"
            )
            return
        return func(context, *args, **kwargs)

    return wrapper


@with_current_task_context
def get(context, key, default=None):
    """Retrieve the value associated to a key in the current task context.

    Fallback to default if the key is missing.
    """
    return context.get(key, default)


@with_current_task_context
def set(context, key, value):
    """Map a key to a value in the current task context.

    If the key is already set, the value is overwritten.
    """
    context[key] = value


@with_current_task_context
def setdefault(context, key, value):
    """If a key is not set in the current task context, map it to a value.

    If the key is already set, its associated value is unchanged.
    """
    context.setdefault(key, value)


@with_current_task_context
def delete(context, key):
    """Delete a key from the current task context."""
    return context.pop(key, None)


def create_task_factory(task_factory=None):
    """Wrap a task factory to support task context inheritance.

    When a new task is spawned, it receives a shallow copy of the parent task
    context. This allows to modify the context of the child task without
    altering the parent one.
    """
    def _default_task_factory(loop, coro):
        """Default function to create a new task."""
        return Task(coro, loop=loop)

    if task_factory is None:
        task_factory = _default_task_factory

    def wrapped(loop, coro):
        parent_task = Task.current_task(loop=loop)
        child_task = task_factory(loop, coro)
        if child_task._source_traceback:
            del child_task._source_traceback[-1]
        context = getattr(parent_task, _TASK_CONTEXT_ATTR, {})
        setattr(child_task, _TASK_CONTEXT_ATTR, dict(context))
        return child_task

    return wrapped


task_factory = create_task_factory()



# -*- coding: utf-8 -*-
# Copyright (c) 2016 - 2020 Sqreen. All rights reserved.
# Please refer to our terms for more information:
#
#     https://www.sqreen.io/terms.html
#
""" lxml strategy to add resolver callbacks
"""
import logging
import sys

from ..._vendors.wrapt import FunctionWrapper
from ...constants import ACTIONS
from ...exceptions import AttackBlocked
from ..hook_point import execute_pre_callbacks
from .import_hook import ImportHookStrategy

LOGGER = logging.getLogger(__name__)


class ResolverCallbackMixin:
    """ lxml resolver called when XML entities are resolved.

    lxml.etree.XMLParser uses a resolver registry which is based on a set data
    structure.  Resolvers can be added and removed but it is not specified in
    what order they are called. In order to be called first, we report a hash
    value of 0 to be inserted as the first element of the hashtable
    underlying the set.
    """

    def __init__(self, strategy):
        self.strategy = strategy

    def __hash__(self):
        return 0

    def resolve(self, *args, **kwargs):
        action = execute_pre_callbacks(
            self.strategy.strategy_key,
            self.strategy,
            None, args, kwargs
        )
        action_status = action.get("status")
        if action_status == ACTIONS["RAISE"]:
            LOGGER.debug(
                "Callback %s detected an attack", action.get("rule_name")
            )
            raise AttackBlocked(action.get("rule_name"))

        elif action_status == ACTIONS["OVERRIDE"]:
            return action.get("new_return_value")


class LXMLResolverStrategy(ImportHookStrategy):
    """ Strategy for adding a resolver to the lxml parser.
    """

    def import_hook_callback(self, original):
        etree = sys.modules["lxml.etree"]

        # the resolver must be a subclass of etree.Resolver otherwise
        # the parser refuses it so we create it at runtime.
        class ResolverCallback(ResolverCallbackMixin, etree.Resolver):
            pass

        resolver = ResolverCallback(self)

        # wrap the function to force add our resolver on the parser.
        def wrapper(wrapped, instance, args, kwargs):
            parser = kwargs.get("parser")
            if parser is None:
                try:
                    parser = etree.get_default_parser()
                except Exception:
                    pass
            resolvers = getattr(parser, "resolvers", None)
            if resolvers is not None:
                resolvers.add(resolver)
                kwargs["parser"] = parser
            return wrapped(*args, **kwargs)

        return FunctionWrapper(original, wrapper)



# -*- coding: utf-8 -*-
# Copyright (c) 2016 - 2020 Sqreen. All rights reserved.
# Please refer to our terms for more information:
#
#     https://www.sqreen.io/terms.html
#
""" Authentication sdk functions
"""
import logging
from datetime import datetime

from ..actions import ACTION_STORE, ActionName
from ..exceptions import ActionBlock, ActionRedirect
from ..runtime_storage import RuntimeStorage, runtime
from ..utils import HAS_TYPING
from . import events

if HAS_TYPING:
    from typing import Mapping, Optional, Union

    UserIdentifier = Union[str, int]
    UserDict = Mapping[str, UserIdentifier]


LOGGER = logging.getLogger(__name__)


def _check_user_action(user_dict):
    # type: (UserDict) -> None
    """Trigger an action for a user when required."""
    action = ACTION_STORE.get_for_user(user_dict)
    if action is None:  # No action planned for this user, do nothing.
        return
    events.track_action(action, {"user": user_dict})
    if action.name == ActionName.BLOCK_USER:
        LOGGER.info("User blocked: %r", user_dict)
        raise ActionBlock(action.iden)
    elif action.name == ActionName.REDIRECT_USER:
        LOGGER.info(
            "User %r is redirected to %r by action %s",
            user_dict,
            action.target_url,
            action.iden,
        )
        raise ActionRedirect(action.iden, action.target_url)


def auth_track(success, **user_identifiers):
    # type: (bool, **UserIdentifier) -> None
    """ Register a successfull or failed authentication attempt (based on
    success boolean) for an user identified by the keyword-arguments.
    For example:

    auth_track(True, email="foobar@example.com") register a successfull
    authentication attempt for user with email "foobar@example.com".

    auth_track(False, user_id=42) register a failed authentication
    attempt for user with id 42.
    """
    _check_user_action(user_identifiers)


def signup_track(**user_identifiers):
    # type: (**UserIdentifier) -> None
    """ Register a new account signup identified by the keyword-arguments.
    For example:

    auth_track(email="foobar@example.com", user_id=42) register
    a new account signup for user identifed by email "foobar@example.com" or
    bu user id 42.
    """
    pass


def identify(user_identifiers, traits=None, storage=runtime):
    # type: (UserDict, Optional[Mapping], RuntimeStorage) -> None
    if traits is None:
        traits = {}
    storage.observe(
        "sdk", ["identify", datetime.utcnow(), user_identifiers, traits], report=False
    )
    _check_user_action(user_identifiers)



# -*- coding: utf-8 -*-
# Copyright (c) 2016 - 2020 Sqreen. All rights reserved.
# Please refer to our terms for more information:
#
#     https://www.sqreen.io/terms.html
#
""" Look for known crawlers user-agents
"""

from logging import getLogger

from .matcher_callback import MatcherRule

LOGGER = getLogger(__name__)


class CrawlerUserAgentMatchesMetricsCB(MatcherRule):
    def pre(self, instance, args, kwargs, **options):
        """ For each request, record an observation with the user_agent without
        case modification.
        """
        request = self.storage.get_current_request()

        if not request:
            LOGGER.warning("No request was recorded abort")
            return

        user_agent = request.client_user_agent

        if not user_agent:
            return

        if self.match(user_agent):
            self.record_observation("crawler", user_agent, 1)



# -*- coding: utf-8 -*-
# Copyright (c) 2016 - 2020 Sqreen. All rights reserved.
# Please refer to our terms for more information:
#
#     https://www.sqreen.io/terms.html
#
""" Config module
"""
import logging
import os
import sys
from os import getenv
from os.path import expanduser, isfile

from .constants import BACKEND_URL

if sys.version_info[0] < 3:
    from ConfigParser import RawConfigParser, Error as ConfigError
else:
    from configparser import RawConfigParser, Error as ConfigError

LOGGER = logging.getLogger(__name__)

HOME_FILE_PATH = expanduser("~/.sqreen.ini")


CONFIG_DEFAULT_VALUE = {
    # The backend URL
    "URL": BACKEND_URL,
    # Default token to use
    "TOKEN": None,
    # If the token is an org token, default app_name
    "APP_NAME": None,
    # Proxy to use if needed
    "PROXY_URL": None,
    # Should the client check the rules signatures
    "RULES_SIGNATURE": True,
    # Custom rules to be used in development
    "RULES": None,
    # Log configuration
    "LOG_LEVEL": "CRITICAL",
    "LOG_LOCATION": None,
    # Should sqreen protection run while application is tested?
    "RUN_IN_TEST": False,
    # Should sqreen report callback performance on newrelic?
    "REPORT_PERF_NR": False,
    # Initial features usually given by the environment
    "INITIAL_FEATURES": "{}",
    # Custom IP headers
    "IP_HEADER": None,
    # Strip http referrer
    "STRIP_HTTP_REFERER": False,
    # Enable PII scrubbing (password, ...)
    "STRIP_SENSITIVE_DATA": True,
    # Sensitive keys to strip
    "STRIP_SENSITIVE_KEYS": "password, secret, passwd, authorization, api_key, apikey, access_token",
    # Sensitive regexp (matches items like credit cards)
    "STRIP_SENSITIVE_REGEX": r"^(?:\d[ -]*?){13,16}$"
}


class Config(object):
    """ Config class to parse several sources of config and merge them

    This try to load configuration by different ways.
    1. Default value
    2. By file, which overrides whatever result we found in 1:
      a. From path in environment variable SQREEN_CONFIG_FILE
      b. From local directory sqreen.ini
      c. From home in ~/.sqreen.ini
    3. From Django configuration directly
    4. From the environment, which overrides whatever result we found in 1 or 2.
    """

    FILE_ENV_VAR = "SQREEN_CONFIG_FILE"
    HOME_FILE_PATH = HOME_FILE_PATH

    def __init__(self, default_values=None):
        self.config = {}

        if default_values is None:
            default_values = CONFIG_DEFAULT_VALUE

        self.default_values = default_values

        self.loaders = [
            self.load_from_default_values,
            self.load_from_file,
            self.load_from_env,
        ]

        self.config_path = None

    def load(self):
        """ Call each loaders and update the config at the end
        """
        base_config = {}

        for loader in self.loaders:
            loaded = loader()

            if loaded:
                base_config.update(loaded)

        self.config = base_config

    def load_from_default_values(self):
        """ Returns default values
        """
        return self.default_values

    def load_from_file(self):
        """ Load from ONE config file, the file path is explained in class
        docstring
        """

        file_path = (
            self.config_path
            or self._file_path_from_env()
            or self._file_path_from_home()
            or self._file_path_from_local()
        )

        if not file_path:
            return {}

        config = RawConfigParser()

        try:
            config.read(file_path)

            config_dict = {}
            for option in config.options("sqreen"):
                config_dict[option.upper()] = config.get("sqreen", option)

            return config_dict
        except ConfigError:
            LOGGER.debug("Error parsing config file %s", file_path)
            return {}

    @staticmethod
    def load_from_env():
        """ Load configuration from os environment variables, variables
        must be prefixed with SQREEN_ to be detected.
        """
        env_config = {}
        for env_var in os.environ:
            if env_var.startswith("SQREEN_"):
                env_config[env_var.replace("SQREEN_", "", 1)] = os.environ[
                    env_var
                ]

        return env_config

    def _file_path_from_env(self):
        """ Return file path if os environement was set and file exists
        """
        path = getenv(self.FILE_ENV_VAR, default=None)

        if path and isfile(path):
            return path

    def _file_path_from_home(self):
        """ Return file path if file exists in home directory
        """
        if self.HOME_FILE_PATH and isfile(self.HOME_FILE_PATH):
            return self.HOME_FILE_PATH

    @staticmethod
    def _file_path_from_local():
        """ Return file path if file exists locally on the project
        """
        if isfile("sqreen.ini"):
            return os.path.join(os.getcwd(), "sqreen.ini")

    def __getitem__(self, name):
        return self.config[name]


CONFIG = Config()
CONFIG.load()



# -*- coding: utf-8 -*-
# Copyright (c) 2016 - 2020 Sqreen. All rights reserved.
# Please refer to our terms for more information:
#
#     https://www.sqreen.io/terms.html
#
""" RSA signature helpers, comes from BackEnd/signature
"""
import json
from abc import ABCMeta
from base64 import b64decode, b64encode
from logging import getLogger

from ._vendors import rsa  # type: ignore
from .utils import is_string

LOGGER = getLogger(__name__)
REQUIRED_KEYS = ("hookpoint", "name", "callbacks", "conditions")
SIGNATURE_VERSION = "rsa_0_1"
PUBLIC_KEY = rsa.PublicKey.load_pkcs1_openssl_pem(
    """-----BEGIN PUBLIC KEY-----
MIIBIjANBgkqhkiG9w0BAQEFAAOCAQ8AMIIBCgKCAQEAvk/Fwmuv1/l90AZe2kvO
+aJsAB0Lobg99urn5MKBLI0JmW6g8NaNW7G+NGgU+4Ro4qQP3zL9B8FXU0RbKX7h
YsAmBHtBwp6wCNeiLVLb8buH8fybJa4bruZmZ1wYN2HPMcSu95GwSIKm+Om3nq84
GjpILgJbN8sGfHzNl7y1cJSj4H6YjAPKmPocJm9wpMb9GVbV/QPgwFPxhFPFijkZ
/rnY4YVDb4lYsYmfUX8JGLSNMDmxtSWCwFsmtWp75p+76Xq0dGGZEKS/CGVwdNfA
tq4p9Kkd4wcZWVWXZNryt2o3kWwt6Oz4dB9EBjjvSBbwlwp2weuCNf8MgMRgoE/9
xwIDAQAB
-----END PUBLIC KEY-----"""
)


class Signature(object):
    """ Signature helpers """

    def __init__(self, digest):
        self.digest = digest

    @staticmethod
    def _normalize_val(val, level):
        """ Normalize a JSON value """

        if level == 0:
            raise Exception("recursion level too deep")

        if isinstance(val, dict):
            val = Signature.normalize(val, None, level - 1)
        elif isinstance(val, list):
            val = [Signature._normalize_val(e, level - 1) for e in val]
            val = "[" + ",".join(val) + "]"
        elif is_string(val) or isinstance(val, int):
            val = json.dumps(val, ensure_ascii=False)
        else:
            raise Exception("JSON hash parsing error (wrong type %r)" % val)
        return val

    @staticmethod
    def _normalize_key(key):
        """ Escapes a JSON key """

        if is_string(key) or isinstance(key, int):
            key = json.dumps(key, ensure_ascii=False)
        else:
            raise Exception(
                "JSON hash parsing error (wrong key type %r)" % key
            )
        return key

    @staticmethod
    def normalize(h, keys=None, level=20):
        """ Normalize the `h`
        Normalization is necessary prior cryptographic signature so that 2 objects have the
        same string representation, independently of their keys order or formatting.
        :param h:     the JSON object to be normalized
        :param keys:  the JSON keys we need to keep
        :param level: the maximum level of recursion we may want to keep
        Return a string that will deterministically represent the JSON.
        """
        """
        Currently a dict { 1: <any>} is serialized as {1:<any>}.
        It is not JSON-valid and cannot be supported in Node.js agent
        The correct form would be {"1":<any>} indifferently for dicts { 1: <any> } or { "1": <any> }
        """

        if level == 0:
            raise Exception("recursion level too deep")
        if not isinstance(h, dict):
            raise Exception("%r is not a dict" % h)

        res = []
        for key in sorted(h):
            v = h[key]
            if keys is not None and key not in keys:
                continue

            v = Signature._normalize_val(v, level - 1)
            k = Signature._normalize_key(key)

            res.append("%s:%s" % (k, v))

        return "{" + ",".join(res) + "}"


class RuleSignature(Signature):
    """ Helpers for signing and verifying rules signatures """

    __metaclass__ = ABCMeta

    def verify_rule(self, hash_rule, keys, version):
        """
        Return True if signature is correct and False otherwise.
        """

        # Encode it in UTF-8.
        value = Signature.normalize(hash_rule, keys).encode("utf-8")

        sig = hash_rule.get("signature")
        if sig is None:
            return False
        ver = sig.get(version)
        if ver is None:
            return False
        encoded_signature = ver["value"]
        signature = b64decode(encoded_signature)

        return self.verify(value, signature)

    def sign_rule(self, hash_rule, keys, version):
        """ Add the signature inside of the hash_rule. """

        # Encode it in UTF-8.
        value = Signature.normalize(hash_rule, keys).encode("utf-8")

        signature = self.sign(value)

        sig_hash = hash_rule.setdefault("signature", {})
        sig_hash.update(
            {version: {"value": b64encode(signature), "keys": keys}}
        )


class RSASigner(RuleSignature):
    """ Generates signatures given a private key using RSA. """

    def __init__(self, private_key, digest=None):
        super(RSASigner, self).__init__(digest)
        self.private_key = rsa.PrivateKey.load_pkcs1(private_key)

    def sign(self, value):
        """ Return the signed value parameter. """
        signature = rsa.sign(value, self.private_key, "SHA-512")
        return signature


class RSAVerifier(RuleSignature):
    """ Allow to verify signature given a public key using RSA. """

    def __init__(self, public_key=None, digest=None):

        super(RSAVerifier, self).__init__(digest)

        if public_key is None:
            self.public_key = PUBLIC_KEY
        else:
            self.public_key = rsa.PublicKey.load_pkcs1_openssl_pem(public_key)

    def verify(self, value, signature):
        """ Return True if value matches the signature, False otherwise.
        """
        try:
            return rsa.verify(value, signature, self.public_key)
        except rsa.VerificationError:
            return False



# -*- coding: utf-8 -*-
# Copyright (c) 2016 - 2020 Sqreen. All rights reserved.
# Please refer to our terms for more information:
#
#     https://www.sqreen.io/terms.html
#
""" Sqreen python agent package
"""
from .__about__ import __author__, __email__, __version__
from .runner_thread import start
from .sdk.auth import auth_track, identify, signup_track
from .sdk.events import track

__all__ = [
    "__author__",
    "__email__",
    "__version__",
    "start",
    "identify",
    "auth_track",
    "signup_track",
    "track",
]



# -*- coding: utf-8 -*-
# Copyright (c) 2016 - 2020 Sqreen. All rights reserved.
# Please refer to our terms for more information:
#
#     https://www.sqreen.io/terms.html
#
""" Remote exception module
"""
import datetime
import sys
import traceback

from .__about__ import __version__


def traceback_formatter(backtrace):
    """ Accept a backtrace in the format of traceback.extract_tb or
    traceback.extract_stack and returns a list of dictionary matching this
    format::

    {
        'file': FILENAME,
        'line_number': LINE_NUMBER,
        'method': FUNCTION_NAME
    }
    """
    frames = []
    for frame in backtrace:
        filename, line_number, function_name, _ = frame
        frames.append(
            {
                "file": filename,
                "line_number": line_number,
                "method": function_name,
            }
        )
    return frames


def raw_traceback_formatter(raw_backtrace):
    """ Accept a traceback object, convert it to a traceback and returns the
    same format than backtrack_formatter.
    """
    return traceback_formatter(traceback.extract_tb(raw_backtrace))


class RemoteException(object):
    def __init__(
        self,
        name,
        msg,
        raw_backtrace=None,
        callback_payload=None,
        exception_payload=None,
        request_payload=None,
        stack=None,
        at=None,
    ):
        self.exception_class = name
        self.exception_msg = msg

        if at is None:
            at = datetime.datetime.utcnow()
        self.at = at

        self.raw_backtrace = raw_backtrace
        if raw_backtrace is not None:
            self.backtrace = raw_traceback_formatter(self.raw_backtrace)
        else:
            self.backtrace = []

        self.stack = None
        if stack is not None:
            self.stack = traceback_formatter(stack)
            self.backtrace = self.stack + self.backtrace

        if callback_payload is None:
            callback_payload = {}
        self.callback_payload = callback_payload

        self.exception_payload = exception_payload
        self.request_payload = request_payload

    @classmethod
    def from_exc_info(cls, exc_info=None, **kwargs):
        """ Create a RemoteException from sys.exc_info.
        """
        if exc_info is None:
            exc_info = sys.exc_info()
        return cls(exc_info[0].__name__, str(exc_info[1]), raw_backtrace=exc_info[2], **kwargs)

    def to_dict(self):
        """ Returns information about exception, backtrace and request merged
        into initial payload
        """
        base_payload = {"infos": {}}

        # Base fields
        base_payload["rule_name"] = self.callback_payload.pop(
            "rule_name", None
        )
        base_payload["rulespack_id"] = self.callback_payload.pop(
            "rulespack_id", None
        )
        base_payload["rule_signature"] = self.callback_payload.pop(
            "rule_signature", None
        )

        if self.callback_payload:
            base_payload["infos"]["callback"] = self.callback_payload

        if self.exception_payload:
            base_payload["infos"]["exception"] = self.exception_payload

        if self.request_payload:
            base_payload.update(self.request_payload)

        base_payload.update(
            {
                "klass": self.exception_class,
                "message": self.exception_msg,
                "context": {"backtrace": self.backtrace},
            }
        )

        return base_payload

    def to_signal(self):
        """ Exception signal.
        """
        signal = dict(
            type="point",
            signal_name="sq.agent.exception",
            payload_schema="exception/2020-01-01T00:00:00.000Z",
        )

        rule_name = self.callback_payload.pop("rule_name", None)
        if rule_name is not None:
            rulespack_id = self.callback_payload.pop(
                "rulespack_id", "0000000000000000000000000000000000000000")
            signal["source"] = "sqreen:rule:{}:{}".format(rulespack_id, rule_name)
        else:
            signal["source"] = "sqreen:agent:{}".format(__version__)

        signal["time"] = self.at
        signal["location"] = {"stack_trace": self.backtrace}

        infos = {}
        if self.callback_payload:
            infos["callback"] = dict(self.callback_payload)

        if self.exception_payload:
            infos["exception"] = self.exception_payload

        signal["payload"] = {
            "klass": self.exception_class,
            "message": self.exception_msg,
            "infos": infos,
        }

        if self.request_payload:
            signal["payload"].update(self.request_payload)

        return signal



# -*- coding: utf-8 -*-
# Copyright (c) 2016 - 2020 Sqreen. All rights reserved.
# Please refer to our terms for more information:
#
#     https://www.sqreen.io/terms.html
#
"""Asynchronous variant of hook_point."""

import logging
import sys

from .._vendors.wrapt import FunctionWrapper
from ..constants import ACTIONS
from ..exceptions import ActionBlock, ActionRedirect, AttackBlocked
from .hook_point import (
    execute_failing_callbacks,
    execute_post_callbacks,
    execute_pre_callbacks,
    valid_args,
)

LOGGER = logging.getLogger(__name__)


def async_hook_point(strategy, hook_name, hook_method, original):
    """Asynchronous variant of hook_point."""
    from asyncio import coroutine

    @coroutine
    def wrapper(wrapped, instance, args, kwargs):
        LOGGER.debug(
            "Checking before async hook point of %s for %s/%s",
            strategy,
            hook_name,
            hook_method,
        )
        strategy.before_hook_point()
        key = (hook_name, hook_method)

        # Call pre callbacks.
        action = execute_pre_callbacks(key, strategy, instance, args, kwargs)

        if action.get("status") == ACTIONS["RAISE"]:
            LOGGER.debug(
                "Callback %s detected an attack", action.get("rule_name")
            )
            raise AttackBlocked(action.get("rule_name"))
        elif action.get("status") == ACTIONS["ACTION_BLOCK"]:
            LOGGER.debug(
                "Action %s blocked the request", action.get("action_id")
            )
            raise ActionBlock(action.get("action_id"))
        elif action.get("status") == ACTIONS["ACTION_REDIRECT"]:
            LOGGER.debug(
                "Action %s redirected the request to %r",
                action.get("action_id"),
                action["target_url"],
            )
            raise ActionRedirect(action.get("action_id"), action["target_url"])
        elif action.get("status") == ACTIONS["OVERRIDE"]:
            return action.get("new_return_value")
        elif action.get("status") == ACTIONS["MODIFY_ARGS"]:
            if valid_args(action["args"]):
                args, kwargs = action["args"]

        # Call the original method.
        retry = True
        while retry is True:
            try:
                retry = False
                # Try to call the original coroutine.
                result = yield from wrapped(*args, **kwargs)
            except Exception:
                # In case of error, call fail callbacks with exception infos.
                exc_info = sys.exc_info()

                # Either raise an exception, set a return value or retry.
                action = execute_failing_callbacks(
                    key, strategy, instance, exc_info, args, kwargs
                )

                if action.get("status") == ACTIONS["RAISE"]:
                    LOGGER.debug(
                        "Callback %s detected an attack",
                        action.get("rule_name"),
                    )
                    raise AttackBlocked(action.get("rule_name"))
                elif action.get("status") == ACTIONS["RETRY"]:
                    retry = True
                elif action.get("status") == ACTIONS["OVERRIDE"]:
                    return action.get("new_return_value")

                # Be sure to raise if no retry or override.
                if retry is False:
                    raise

        # Then call post callback in reverse order to simulate decorator
        # behavior.
        action = execute_post_callbacks(
            key, strategy, instance, result, args, kwargs
        )

        if action.get("status") == ACTIONS["RAISE"]:
            LOGGER.debug(
                "Callback %s detected an attack", action.get("rule_name")
            )
            raise AttackBlocked(action.get("rule_name"))
        elif action.get("status") == ACTIONS["OVERRIDE"]:
            return action.get("new_return_value")

        # Return the original value.
        return result

    return FunctionWrapper(original, wrapper)



# -*- coding: utf-8 -*-
# Copyright (c) 2016 - 2020 Sqreen. All rights reserved.
# Please refer to our terms for more information:
#
#     https://www.sqreen.io/terms.html
#

import hashlib
import logging
import uuid
from threading import Lock
from weakref import WeakKeyDictionary

from ..binding_accessor import BindingAccessor
from ..exceptions import InvalidArgument, SqreenException
from ..rules import RuleCallback
from ..runner import BaseRunner
from ..utils import HAS_TYPING, is_unicode

if HAS_TYPING:
    from typing import Mapping, MutableMapping
else:
    from collections import Mapping

try:
    from sq_native import waf, input as waf_input  # type: ignore
except ImportError:
    waf = None  # pragma: no cover


LOGGER = logging.getLogger(__name__)


class WAFException(SqreenException):

    def __init__(self, msg, waf_rules_id, ret, cbargs, arguments):
        super(WAFException, self).__init__(msg)
        self.waf_rules_id = waf_rules_id
        self.ret = ret
        self.cbargs = cbargs
        self.arguments = arguments

    def exception_infos(self):
        data = self.ret.data
        if data is not None and isinstance(data, bytes):
            data = data.decode("utf-8")
        return {
            "args": self.cbargs,
            "waf": {
                "args": self.arguments,
                "waf_rule": self.waf_rules_id,
                "error_code": self.ret.action,
                "error_data": data,
            }
        }


class WAFInstance:
    """
    Per runner WAF instance.
    """

    def __init__(self, waf_rules_id, waf_rules):
        unique_id = "{}_{}".format(uuid.uuid4(), waf_rules_id)
        ret = waf.initialize(unique_id, waf_rules)
        if ret is not True:
            msg = "failed to initialize the WAF rules"
            raise InvalidArgument(msg)
        self.unique_id = unique_id
        self._waf_rules_id = waf_rules_id

    def run(self, params, budget_ms):
        """ Run the WAF on the parameters.
        """
        return waf.run(self.unique_id, params, int(budget_ms * 1000))

    def __del__(self):
        unique_id = getattr(self, "unique_id", None)
        if unique_id is not None:
            waf.clear(unique_id)

    class DummyRunner(BaseRunner):
        pass

    _pool = WeakKeyDictionary()  # type: MutableMapping[BaseRunner, WAFInstance]
    _lock = Lock()
    _dummy_runner = DummyRunner()

    @classmethod
    def get_for_runner(cls, waf_rules_name, waf_rules, runner=None):
        """ Get the WAF instance associated with the runner.
        """
        if runner is None:
            runner = cls._dummy_runner

        if is_unicode(waf_rules):
            waf_rules = waf_rules.encode("utf-8")
        if not isinstance(waf_rules, bytes):
            raise InvalidArgument("Invalid WAF rules")

        waf_rules_hash = hashlib.sha256(waf_rules).hexdigest()
        waf_rules_id = "{}_{}".format(waf_rules_hash, waf_rules_name)

        with cls._lock:
            instances = cls._pool.get(runner)
            if instances is None:
                instances = cls._pool[runner] = {}
            instance = instances.get(waf_rules_id)
            if instance is None or instance._waf_rules_id != waf_rules_id:
                LOGGER.debug("Instanciate new WAF context for %s", waf_rules_id)
                instance = instances[waf_rules_id] = cls(
                    waf_rules_id, waf_rules)
            return instance


class WAFCB(RuleCallback):
    """ WAF Callback
    """

    SUPPORTS_BUDGET = True
    DEFAULT_MAX_BUDGET_MS = 5
    DEFAULT_MAX_INPUT_DEPTH = 10
    DEFAULT_MAX_STRING_LENGTH = 4096
    DEFAULT_MAX_INPUT_ITEMS = 150

    def __init__(self, *args, **kwargs):
        super(WAFCB, self).__init__(*args, **kwargs)
        if not self.module_is_available():
            msg = "WAF is disabled because the native module is not installed"
            raise InvalidArgument(msg)

        values = self.data.get("values", {})
        waf_rules = values.get("waf_rules")

        self._inst = WAFInstance.get_for_runner(self.rule_name, waf_rules, self.runner)

        bas = {}
        for item in values.get("binding_accessors", []):
            if isinstance(item, Mapping):
                bas[item["ba"]] = (BindingAccessor(item["ba"]), item.get("default", None))
            else:
                bas[item] = (BindingAccessor(item), None)
        self.binding_accessors = bas
        self.max_budget_ms = values.get("max_budget_ms", self.DEFAULT_MAX_BUDGET_MS)
        self.max_input_depth = values.get("max_input_depth", self.DEFAULT_MAX_INPUT_DEPTH)
        self.max_string_length = values.get("max_string_length", self.DEFAULT_MAX_STRING_LENGTH)
        self.max_input_items = values.get("max_input_items", self.DEFAULT_MAX_INPUT_ITEMS)

    def pre(self, instance, args, kwargs, **options):
        request = self.storage.get_current_request()
        if request is None:
            return

        budget = self.get_remaining_budget(options)
        if budget is None:
            budget = self.max_budget_ms
        elif budget <= 0:
            LOGGER.debug("skipping WAF because budget is negative")
            return

        binding_eval_args = {
            "request": request,
            "inst": instance,
            "args": self.storage.get_current_args(args),
            "kwargs": kwargs,
            "data": self.data,
        }

        params = {}
        for expr, (ba, default) in self.binding_accessors.items():
            with self.storage.trace() as trace:
                try:
                    res = ba.resolve(**binding_eval_args)
                    params[expr] = default if res is None else res
                except (ValueError, KeyError, AttributeError):
                    params[expr] = default
            budget -= trace.duration_ms
            if budget <= 0:
                LOGGER.debug("skipping WAF because %r evaluation is overtime", expr)
                return

        LOGGER.debug("WAF parameters: %r", params)
        with self.storage.trace() as trace:
            pw_params = waf_input.PWArgs.from_python(
                params, max_depth=self.max_input_depth,
                max_string_length=self.max_string_length,
                max_items=self.max_input_items)

        budget -= trace.duration_ms
        if budget <= 0:
            LOGGER.debug("skipping WAF because params convertion is overtime")
            return

        ret = self._inst.run(pw_params, min(budget, self.max_budget_ms))
        LOGGER.debug("WAF returned: %r", ret)
        if ret.action in (waf.PW_BLOCK, waf.PW_MONITOR):
            self.record_attack({
                "waf_data": ret.data,
                # This is a hack to detect if we are running inside the PHP Daemon
                "ba_type": "php-daemon" if hasattr(self.runner, "ident") else "python"
            })
            if ret.action == waf.PW_BLOCK and self.block is True:
                return {"status": "raise", "rule_name": self.rule_name}
        elif ret.action < 0 and ret.action != -5:  # -5 is WAF timeout
            raise WAFException(
                "WAF returned unexpected action", self._inst.unique_id, ret, args, params)

    @staticmethod
    def module_is_available():
        """ Return True if the native module is present, False otherwise.
        """
        return waf is not None



# -*- coding: utf-8 -*-
# Copyright (c) 2016 - 2020 Sqreen. All rights reserved.
# Please refer to our terms for more information:
#
#     https://www.sqreen.io/terms.html
#

import logging

from .runtime_infos import RuntimeInfos

LOGGER = logging.getLogger(__name__)


class PayloadCreator(object):
    """Create attack payloads."""

    SECTIONS = ("request", "response", "params", "headers", "local")

    @classmethod
    def get_payload(cls, request, response=None, sections=None):
        if sections is None:
            sections = cls.SECTIONS
        payload = {}
        if "request" in sections:
            payload["request"] = request.request_payload
            if "client_ip" in request.request_payload:
                payload["client_ip"] = request.request_payload["client_ip"]
        if "response" in sections and response is not None:
            payload["response"] = response.response_payload
        if "params" in sections:
            # Don't send cookies and messages.
            payload["params"] = {
                "form": request.form_params,
                "query": request.query_params,
                "other": request.view_params,
                "json": request.json_params,
            }
        if "headers" in sections:
            payload["headers"] = request.get_client_ips_headers()
        if "local" in sections:
            payload["local"] = RuntimeInfos.local_infos()
        return payload



# -*- coding: utf-8 -*-
# Copyright (c) 2016 - 2020 Sqreen. All rights reserved.
# Please refer to our terms for more information:
#
#     https://www.sqreen.io/terms.html
#
""" Base Request class
"""
import sys
import traceback
import uuid
from collections import defaultdict

from ..config import CONFIG
from ..remote_exception import traceback_formatter
from ..utils import HAS_TYPING, flatten, to_unicode_safe
from .ip_utils import get_real_user_ip

if sys.version_info[0] >= 3:
    from urllib.parse import unquote_to_bytes
else:
    from urllib import unquote as unquote_to_bytes

if HAS_TYPING:
    from typing import List


# Header name that we need to sends back on attack in order to compute the
# real user ip
DEFAULT_CLIENT_IP_HEADERS = (
    "HTTP_X_FORWARDED_FOR",
    "X_FORWARDED_FOR",
    "HTTP_X_REAL_IP",
    "HTTP_CLIENT_IP",
    "HTTP_X_FORWARDED",
    "HTTP_X_CLUSTER_CLIENT_IP",
    "HTTP_FORWARDED_FOR",
    "HTTP_FORWARDED",
    "HTTP_VIA",
    "REMOTE_ADDR",
)
CLIENT_IP_HEADERS = []  # type: List[str]


def load_config_ip_header():
    del CLIENT_IP_HEADERS[:]
    if CONFIG["IP_HEADER"]:
        CLIENT_IP_HEADERS.append(
            "HTTP_{}".format(CONFIG["IP_HEADER"].upper().replace("-", "_")))
    CLIENT_IP_HEADERS.extend(DEFAULT_CLIENT_IP_HEADERS)


load_config_ip_header()


class BaseRequest(object):
    """ Base request class common to all request wrapper classes.
    """

    def __init__(self, storage=None):
        self.storage = storage

        self._request_headers = None
        self._request_id = None

        # Used by params_contains method in the reflected_xss callback
        # XXX: could be replaced by binding accessor resolutions, but they are
        # less efficient as we would recompute flat_values once per callback.
        self._flat_request_params = None
        self._flat_request_params_keys = None

    @property
    def remote_addr(self):
        raise NotImplementedError

    @property
    def hostname(self):
        raise NotImplementedError

    @property
    def method(self):
        raise NotImplementedError

    @property
    def referer(self):
        raise NotImplementedError

    @property
    def client_user_agent(self):
        raise NotImplementedError

    @property
    def path(self):
        raise NotImplementedError

    @property
    def scheme(self):
        raise NotImplementedError

    @property
    def server_port(self):
        raise NotImplementedError

    @property
    def remote_port(self):
        raise NotImplementedError

    @property
    def raw_headers(self):
        raise NotImplementedError

    @property
    def form_params(self):
        raise NotImplementedError

    @property
    def query_params(self):
        raise NotImplementedError

    @property
    def view_params(self):
        raise NotImplementedError

    @property
    def json_params(self):
        raise NotImplementedError

    @property
    def cookies_params(self):
        raise NotImplementedError

    @property
    def body(self):
        raise NotImplementedError

    @property
    def request_id(self):
        if self._request_id is None:
            self._request_id = uuid.uuid4().hex
        return self._request_id

    @property
    def request_uri(self):
        uri = [self.path]
        query_string = self.get_raw_header("QUERY_STRING")
        if query_string:
            uri.append("?")
            query_string = query_string.replace('+', ' ')
            if hasattr(query_string, "encode"):
                query_string = query_string.encode()
            uri.append(to_unicode_safe(unquote_to_bytes(query_string)))
        return "".join(uri)

    @property
    def request_payload(self):
        """ Returns current request payload with the backend expected field
        name. All fields should be serializable to JSON.
        """
        if CONFIG["STRIP_HTTP_REFERER"]:
            referer = None
        else:
            referer = self.referer

        return {
            "rid": self.request_id,
            "remote_ip": self.remote_addr,
            "client_ip": self.client_ip,
            "host": self.hostname,
            "verb": self.method,
            "referer": referer,
            "user_agent": self.client_user_agent,
            "path": self.path,
            "scheme": self.scheme,
            "port": self.server_port,
            "remote_port": self.remote_port,
        }

    @property
    def request_params(self):
        """ Returns all the inputs that are controllable by the user.
        All fields should be serializable to JSON.
        """
        return {
            "form": self.form_params,
            "query": self.query_params,
            "other": self.view_params,
            "json": self.json_params,
            "cookies": self.cookies_params,
        }

    @property
    def request_params_list(self):
        """ Returns all the inputs that are controllable by the user without
        keys. All fields should be serializable to JSON.
        """
        return [
            self.form_params,
            self.query_params,
            self.view_params,
            self.json_params,
            self.cookies_params,
        ]

    @property
    def request_params_filtered(self):
        """ Returns all the inputs that are controllable by the user without cookies.
        This is based on the Ruby reference implementation.
        """
        ret = self.request_params
        ret.pop("cookies", None)
        return ret

    def _flatten_request_params(self):
        """ Flatten the request params to compute the values and keys for
        quick access
        """
        keys, values = flatten(self.request_params_list)
        self._flat_request_params_keys = set(keys)
        self._flat_request_params = values

    @property
    def flat_request_params(self):
        """ Return only the request params values for all request params
        """
        if self._flat_request_params is None:
            self._flatten_request_params()
        return self._flat_request_params

    @property
    def flat_request_params_keys(self):
        """ Return only the request params names for all request params
        """
        if self._flat_request_params_keys is None:
            self._flatten_request_params()
        return self._flat_request_params_keys

    def params_contains(self, param):
        """ Return True if the parameter given in input is present in the
        request inputs.
        """
        return param in self.flat_request_params

    @property
    def request_headers(self):
        if self._request_headers is None:
            self._request_headers = defaultdict(lambda: None)

            for key, value in self.raw_headers.items():
                key = to_unicode_safe(key)
                value = to_unicode_safe(value)
                if not key.startswith("HTTP_"):
                    if key == "CONTENT_LENGTH" and value:
                        # Some WSGI implementations set CONTENT_LENGTH to
                        # an empty string when not set in the HTTP request.
                        # We actually prefer to consider the header not set.
                        pass
                    elif key != "CONTENT_TYPE":
                        continue
                self._request_headers[key] = value
        return self._request_headers

    @property
    def content_type(self):
        return self.get_raw_header("CONTENT_TYPE")

    @property
    def content_length(self):
        value = self.get_raw_header("CONTENT_LENGTH")
        if value is not None:
            try:
                return int(value)
            except ValueError:
                pass
        return None

    @property
    def caller(self):
        return traceback.format_stack()

    @property
    def raw_caller(self):
        return traceback_formatter(traceback.extract_stack())

    @property
    def client_ip(self):
        """String formatted client IP address."""
        ip = self.raw_client_ip
        if ip is not None:
            return to_unicode_safe(ip)

    @property
    def raw_client_ip(self):
        """Client IP address."""
        return get_real_user_ip(self.remote_addr, *self.iter_client_ips())

    def get_raw_header(self, name, default=None):
        """ Get a specific raw header."""
        return self.raw_headers.get(name, default)

    def iter_client_ips(self):
        """Yield the client IPs set in raw headers."""
        for header_name in CLIENT_IP_HEADERS:
            value = self.get_raw_header(header_name)
            if value:
                yield value

    def get_client_ips_headers(self):
        """ Return raw headers that can be used to find the real client ip
        """
        headers = []
        for header_name in CLIENT_IP_HEADERS:
            value = self.get_raw_header(header_name)
            if value:
                headers.append([header_name, to_unicode_safe(value)])

        return headers


class BaseResponse(object):

    @property
    def status_code(self):
        raise NotImplementedError

    @property
    def content_type(self):
        raise NotImplementedError

    def content_length(self):
        raise NotImplementedError

    @property
    def response_payload(self):
        """Returns the response information understood by the backend.
        """
        return {
            "status": self.status_code,
            "content_type": self.content_type,
            "content_length": self.content_length,
        }



# -*- coding: utf-8 -*-
# Copyright (c) 2016 - 2020 Sqreen. All rights reserved.
# Please refer to our terms for more information:
#
#     https://www.sqreen.io/terms.html
#
""" Deliverer classes package
"""
from .batch import BatchDeliverer
from .signal import SignalDeliverer
from .simple import SimpleDeliverer

__all__ = ["SimpleDeliverer", "BatchDeliverer", "SignalDeliverer"]


def get_deliverer(batch_size, max_staleness, session, use_signals=False):
    """ Helper function to returns the correct deliverer class for the
    batch_size and max_stalennes parameters
    """
    if batch_size < 1:
        return SimpleDeliverer(session)
    elif use_signals:
        return SignalDeliverer(session, batch_size, max_staleness)
    else:
        return BatchDeliverer(session, batch_size, max_staleness)
